<?php
/**
 * Cache class
 *
 * @package WavePlayer
 */

namespace PerfectPeach\WavePlayer;

defined( 'ABSPATH' ) || exit;

/**
 * Cache providing methods to handle the caching of data across the plugin.
 *
 * @since 3.7.0
 */
class Cache {
	/**
	 * Get transient
	 *
	 * @param string $type The type of transient being stored: either 'instance_data' or 'preview_files'.
	 * @param string $key  The key of the transient being stored:
	 *                     the hash of an instance configuration if type is 'instance_data'
	 *                     or the product ID if type is 'preview_files'.
	 * @return mixed
	 */
	public static function get_transient( $type, $key ) {
		return get_transient( "waveplayer_{$type}_{$key}" );
	}

	/**
	 * Set transient
	 *
	 * @param string $type       The type of transient being stored: either 'instance_data' or 'preview_files'.
	 * @param string $key        The key of the transient being stored:
	 *                           the hash of an instance configuration if type is 'instance_data'
	 *                           or the product ID if type is 'preview_files'.
	 * @param mixed  $value      The value being stored in the transient.
	 * @param int    $expiration The expiration time, in seconds.
	 * @return void
	 */
	public static function set_transient( $type, $key, $value, $expiration = null ) {
		if ( is_null( $expiration ) ) {
			$expiration = WEEK_IN_SECONDS;
		}

		$transients = get_transient( "waveplayer_{$type}" );

		if ( ! $transients ) {
			$transients = array();
		}

		$transients[] = $key;
		set_transient( "waveplayer_{$type}", $transients, $expiration );
		set_transient( "waveplayer_{$type}_{$key}", $value, $expiration );
	}

	/**
	 * Delete transient
	 *
	 * @param string $type The type of transient being stored: either 'instance_data' or 'preview_files'.
	 * @param string $key  The key of the transient being stored:
	 *                     the hash of an instance configuration if type is 'instance_data'
	 *                     or the product ID if type is 'preview_files'.
	 * @return void
	 */
	public static function delete_transient( $type, $key ) {
		$transients = get_transient( "waveplayer_{$type}" );

		if ( $transients && isset( $transients[ $key ] ) ) {
			unset( $transients[ $key ] );
			set_transient( "waveplayer_{$type}", $transients );
		}

		delete_transient( "waveplayer_{$type}_{$key}" );
	}

	/**
	 * Delete all transients
	 *
	 * @param string $type The type of transient being stored: either 'instance_data' or 'preview_files'
	 *                     Optional: if not provided, all transients will be deleted.
	 * @return int The number of transients being deleted.
	 */
	public static function delete_transients( $type = '' ) {
		$types = array(
			'global',
			'instance_data',
			'preview_files',
		);

		if ( $type ) {
			$types = array_intersect( $types, array( $type ) );
		}

		$count = 0;

		foreach ( $types as $current_type ) {
			$transients = get_transient( "waveplayer_{$current_type}" );

			if ( $transients ) {
				foreach ( $transients as $key ) {
					if ( delete_transient( "waveplayer_{$current_type}_{$key}" ) ) {
						$count++;
					}
				}
			}

			delete_transient( $current_type );
		}

		return $count;
	}
}
